#include "config.h"
#include "afflib.h"
#include "afflib_i.h"
#include "utils.h"
#ifdef HAVE_ERR_H
#include "err.h"
#endif

#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#endif



/* Given an AFFILE, return a seglist.
 * Returns -1 if failure, 0 if success.
 */
int seglist::get_seglist(AFFILE *af)
{
    if(af_rewind_seg(af)) return -1;
    char name_[AF_MAX_NAME_LEN];
    size_t len_=0;
    u_long arg_=0;
    while(af_get_next_seg(af,name_,sizeof(name_),&arg_,0,&len_)==0){
	// We shouldn't have 0-len segment names, but we do in some files.
	// Don't copy these segments.
	if(strlen(name_)>0){		
	    seginfo si(name_,len_,arg_);
	    push_back(si);
	}
    }
    return 0;
}


bool ends_with(string str,string ending)
{
    if(str.length() < ending.length()) return false;
    if(str.substr(str.length()-ending.length()) == ending) return true;
    return false;
}

int parse_chain(const string &name)
{
    char ch;
    int num;
    if(sscanf(name.c_str(),AF_CUSTODY_CHAIN_SEG"%c",&num,&ch)==1) return num;
    return -1;
}

int highest_chain(seglist &segments)
{
    int  highest_chain   = -1;
    for(seglist::const_iterator seg = segments.begin(); seg!=segments.end() ;seg++){
	/* Are any of the segments in the input file signed?
	 * If so, we can't use the AFFLIB signing mechanisms, only the
	 * chain-of-custody mechanisms in this program.
	 */
	/* Are there any chain of custody segments? */
	int num= parse_chain(seg->name);
	if(num>highest_chain) highest_chain = num;
    }
    return highest_chain;

}

#ifdef HAVE_OPENSSL_BIO_H
/* BIO_xmlescape:
 * sends str to the bio, escaping for XML.
 */
int BIO_write_xml_escape(BIO *bio,const char *str,int parsed)
{
    while(*str){
	switch(*str){
	case '&': BIO_write(bio,"&amp;",5);break;
	case '<': BIO_write(bio,"&lt;",4);break;
	case '>': BIO_write(bio,"&gt;",4);break;
	case '"': BIO_write(bio,"&quot;",6);break;
	case '\'': BIO_write(bio,"&apos;",6);break;
	case '\\':
	    if(parsed) BIO_write(bio,str,1);
	    else BIO_write(bio,"\\\\",2);
	    break;
	default: BIO_write(bio,str,1);break;
	}
	str++;
    }
    return 0;
}
#endif

#ifdef USE_AFFSIGS
char *aff_bom::get_notes()
{
    if(isatty(fileno(stdin))){
	printf("Enter notes. Terminate input with a '.' on a line by itself:\n");
    }
    if(notes) return notes;
    notes = (char *)calloc(1,1);
    while(notes){
	char buf2[1024];
	char *val=0;
	
#ifdef HAVE_LIBREADLINE
	if(isatty(fileno(stdin))){
	    val = readline("");
	}
#endif
	if(val==0){
	    memset(buf2,0,sizeof(buf2));
	    val = fgets(buf2,sizeof(buf2)-1,stdin);
	    if(val==0) break;
	}
	if(strcmp(val,".")==0) break;
	notes = (char *)realloc(notes,strlen(notes)+strlen(val)+1);
	strcat(notes,val);
    }
    return notes;
}




#ifdef HAVE_OPENSSL_BIO_H
int aff_bom::read_files(const char *cert_file,const char *key_file)
{
    BIO *bp_cert = BIO_new_file(cert_file,"r"); // read the certfile
    PEM_read_bio_X509(bp_cert,&cert,0,0); // get an x509 cert
    BIO_free(bp_cert);
    if(!cert) return -1;		// can't read certificate file
	
    /* Now read the private key */
    BIO *bp_privkey = BIO_new_file(key_file,"r");
    privkey = PEM_read_bio_PrivateKey(bp_privkey,0,0,0);
    BIO_free(bp_privkey);
    if(privkey==0){
	X509_free(cert);
	cert = 0;
	return -1;
    }
	
    xml = BIO_new(BIO_s_mem());	// where we are writing
    time_t clock = time(0);
    struct tm *tm = localtime(&clock);
    char buf[1024];
    strftime(buf,sizeof(buf),"<date type='ISO 8601'>%FT%T</date>",tm);
    
    BIO_printf(xml,"<%s>\n",AF_XML_CUSTODY_CHAIN);
    BIO_printf(xml,"  %s\n",buf);
    BIO_printf(xml,"  <program>afcopy</program>\n");
    if(opt_note){
	BIO_printf(xml,"  <notes>");
	BIO_write_xml_escape(xml,get_notes(),0);
	BIO_printf(xml,"  </notes>\n");
    }
    BIO_printf(xml,"  <signingcertificate>\n");
    PEM_write_bio_X509(xml,cert);
    BIO_printf(xml,"  </signingcertificate>\n");
    BIO_printf(xml,"  <affsegments>\n");
    return 0;
}

/* Add to the Bill of Materials */
void aff_bom::add(const char *segname,int sigmode,const u_char *seghash,size_t seghash_len)
{
    BIO_printf(xml,"<%s segname='%s' sigmode='%d' alg='sha256'>\n",
	       AF_XML_SEGMENT_HASH,segname,sigmode);
    BIO_flush(xml);
    BIO *b64 = BIO_new(BIO_f_base64());
    xml = BIO_push(b64,xml);
    BIO_write(xml,seghash,seghash_len);
    BIO_flush(xml);
    xml = BIO_pop(b64);
    BIO_printf(xml,"</%s>\n",AF_XML_SEGMENT_HASH);
}

void aff_bom::close()
{
#ifdef HAVE_EVP_SHA256
    /* Terminate the XML block*/
    BIO_printf(xml,"</affsegments>\n");
    BIO_printf(xml,"</%s>\n",AF_XML_CUSTODY_CHAIN);

    /* now sign the XML */
    char *xbuf=0;
    size_t xlen = BIO_get_mem_data(xml,&xbuf);
    unsigned char sig[1024];
    u_int  siglen = sizeof(sig);
    
    EVP_MD_CTX md;
    EVP_SignInit(&md,EVP_sha256());
    EVP_SignUpdate(&md,xbuf,xlen);
    EVP_SignFinal(&md,sig,&siglen,privkey);
    
    /* Write the signature in base64 encoding... */
    BIO *b64 = BIO_new(BIO_f_base64());
    xml = BIO_push(b64,xml);
    BIO_write(xml,sig,siglen);
    BIO_flush(xml);
	
    /* Remove the base64 bio */
    xml = BIO_pop(b64);
#else
    errx(1,"Cannot sign BoM; EVP_sha256 not available");
#endif    
}

int  aff_bom::write(AFFILE *af,seglist &segments)
{
    char segname[AF_MAX_NAME_LEN];
    snprintf(segname,sizeof(segname),AF_CUSTODY_CHAIN_SEG,highest_chain(segments)+1);
    return af_update_seg_frombio(af,segname,0,xml);
}
#endif
#endif
